from dataclasses import dataclass
from datetime import timedelta
from http.client import HTTPResponse
from logging import Logger
from socket import EAI_AGAIN, EAI_FAIL, gaierror
from typing import AsyncContextManager, Callable, Iterable, Optional, Type
from urllib.error import HTTPError
from urllib.request import Request, urlopen
from uuid import UUID
import importlib.metadata
import re

from bs4 import BeautifulSoup
from trio import CancelScope, MemoryReceiveChannel, MemorySendChannel, move_on_at
import trio

from .abc import DeleteEvent, EventStream, Notifier, NotifyInfo, ThreadTitleProvider, UpdateChecker, UpdateEvent


_USER_AGENT = "any:net.jcornell.mentionbot:v{} (by /u/jaxklax)".format(
	importlib.metadata.version(__package__)
)


@dataclass
class TokenRef:
	"""Mutable container for an access token."""

	token: str

	def get_common_headers(self) -> dict[str, str]:
		return {
			"User-Agent": _USER_AGENT,
			"Authorization": "Bearer {}".format(self.token)
		}


def try_request(
	request: Request,
	logger: Logger,
	suppress_codes: list[int] = [404, 429, 500, 503],
) -> Optional[HTTPResponse]:

	try:
		return urlopen(request)
	except gaierror as e:
		if e.errno in [EAI_FAIL, EAI_AGAIN]:
			logger.error(f"DNS failure: {e}")
			return None
		else:
			raise
	except HTTPError as e:
		e.read()
		e.close()
		if e.code in suppress_codes:
			logger.error(f"HTTP {e.code}")
			return None
		else:
			raise


@dataclass
class WsListenerState:
	"""The WebSockets listener uses this object to communicate with the API poll and message sender tasks."""

	# The API poll task uses this scope to force a refresh of the WS connection if it's been too quiet.
	scope: Optional[CancelScope] = None

	# The WS task uses this to track the UUID timestamp of the most recent update message update, so the API poll task
	# can detect when the connection goes silent.
	last_update_ts: Optional[int] = None

	# The WS task sets this when the WS connection goes down and clears it when it's recovered.
	down: bool = True

	# The WS task sets this to the UUID timestamp of the first update recieved on connection recovery so the message
	# sender task knows when to stop checking for deletion.
	resync_ts: Optional[int] = None


def _parse_mentions(body: BeautifulSoup) -> set[str]:
	"""Extract the names of users tagged in an update."""
	return {
		el["href"].removeprefix("/u/")
		for el in body.find_all("a", href = re.compile(r"^/u/[\w-]{3,20}$", re.ASCII))
	}


async def ws_listener_impl(
	event_stream_factory: Callable[[], AsyncContextManager[EventStream]],
	event_stream_catch: Iterable[Type[BaseException]],
	state: WsListenerState,
	event_tx: MemorySendChannel,  # message type `Event'
	logger: Logger,
) -> None:

	while True:
		state.scope = CancelScope()
		with state.scope:
			logger.debug("opening a new connection")
			try:
				async with event_stream_factory() as event_stream:
					first_update = True
					while True:
						event = await event_stream.get_event()
						if isinstance(event, UpdateEvent):
							update_id = UUID(event.payload_data["id"])
							state.last_update_ts = update_id.time
							if first_update:
								# first update after discontinuity
								state.down = False
								state.resync_ts = update_id.time
								first_update = False
						elif isinstance(event, DeleteEvent):
							logger.debug("delete event for " + event.payload)
						else:
							raise RuntimeError(f"expected Event, got {event}")

						await event_tx.send(event)
			except tuple(event_stream_catch) as e:
				logger.warning(f"WebSockets error: {e}")

		state.down = True


async def message_sender_impl(
	notifier: Notifier,
	title_provider: ThreadTitleProvider,
	update_checker: UpdateChecker,
	ws_state: WsListenerState,
	thread_id: str,
	event_rx: MemoryReceiveChannel,  # message type `Event'
	buffer_time: timedelta,
	mention_limit: int,
	logger: Logger,
) -> None:

	@dataclass
	class QueueEntry:
		payload_data: dict
		mentions: set[str]
		arrival: float  # Trio time

	queue: list[QueueEntry] = []

	while True:
		logger.debug("queue: {}".format([e.payload_data["id"] for e in queue]))
		if queue:
			scope = move_on_at(queue[0].arrival + buffer_time.total_seconds())
		else:
			scope = CancelScope()  # no timeout

		with scope:
			event = await event_rx.receive()

		if scope.cancelled_caught:
			# timed out
			entry = queue.pop(0)
			update_id = entry.payload_data["id"]
			if ws_state.down or ws_state.resync_ts and UUID(update_id).time < ws_state.resync_ts:
				notify = update_checker.exists(thread_id, update_id)
			else:
				notify = True

			if notify:
				thread_title = title_provider.title_for(thread_id)
				for recipient in entry.mentions:
					notifier.notify(NotifyInfo(
						recipient, thread_id, thread_title, update_id, entry.payload_data["author"],
						entry.payload_data["body"]
					))
		else:
			# got an event
			if isinstance(event, UpdateEvent):
				update_id = event.payload_data["id"]
				if not any(entry.payload_data["id"] == update_id for entry in queue):
					mentions = _parse_mentions(BeautifulSoup(event.payload_data["body_html"], "html.parser"))
					if mentions:
						if len(mentions) > mention_limit:
							logger.info(f"ignoring {update_id} due to too many mentions ({len(mentions)})")
						else:
							queue.append(QueueEntry(event.payload_data, mentions, trio.current_time()))
			elif isinstance(event, DeleteEvent):
				logger.debug("delete event for " + event.payload)
				queue = [entry for entry in queue if entry.payload_data["name"] != event.payload]
			else:
				raise RuntimeError(f"invalid event type {type(event)}")
