"""
Mentionbot: Notify users by private message when they're /u/-mentioned in a live thread.
"""

from argparse import ArgumentParser
from configparser import ConfigParser
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import timedelta
from logging import Formatter, getLogger, Logger, NOTSET, StreamHandler
from sys import stdout
from typing import ClassVar, Iterator, Optional, Type
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen
from uuid import UUID
import json
import logging

from trio import MemorySendChannel, open_memory_channel, open_nursery
from trio_websocket import ConnectionClosed, HandshakeError, open_websocket_url, WebSocketConnection
import psycopg2
import trio

from . import TokenRef, message_sender_impl, try_request, ws_listener_impl, WsListenerState
from .abc import DeleteEvent, Event, EventStream, Notifier, NotifyInfo, ThreadTitleProvider, UpdateChecker, UpdateEvent


async def _token_update_impl(period: timedelta, db_conn, auth_id: int, task_status) -> None:
	def fetch_token() -> str:
		cursor = db_conn.cursor()
		cursor.execute(
			"""
				select access_token from public.reddit_app_authorization
				where id = %s
			""",
			(auth_id,)
		)
		[(token,)] = cursor.fetchall()
		return token

	ref = TokenRef(fetch_token())
	task_status.started(ref)
	while True:
		await trio.sleep(period.total_seconds())
		ref.token = fetch_token()


async def _api_poll_impl(
	ws_state: WsListenerState,
	token_ref: TokenRef,
	thread_id: str,
	event_tx: MemorySendChannel,  # message type `Event'
	poll_period: timedelta,
	max_ws_delay: timedelta,
	logger: Logger,
) -> None:

	base_url = f"https://oauth.reddit.com/live/{thread_id}"

	latest_payload: Optional[dict] = None
	while True:
		# fetch updates and emit events
		params = {
			"raw_json": "1",
		}
		if latest_payload:
			updates = []
			while True:
				params["before"] = latest_payload["data"]["name"]
				url = base_url + "?" + urlencode(params)
				resp = try_request(Request(url, headers = token_ref.get_common_headers()), logger)
				if resp:
					with resp:
						children = json.load(resp)["data"]["children"]
					if children:
						updates.extend(reversed(children))
						latest_payload = children[0]
					else:
						break
				else:
					break
		else:
			params["limit"] = "1"
			url = base_url + "?" + urlencode(params)
			resp = try_request(Request(url, headers = token_ref.get_common_headers()), logger)
			if resp:
				with resp:
					updates = json.load(resp)["data"]["children"]
				if updates:
					latest_payload = updates[0]
			else:
				updates = []

		for update in updates:
			await event_tx.send(UpdateEvent(update["data"]))

		# check if WS connection is behind
		if not ws_state.down and latest_payload:
			max_delta = max_ws_delay.total_seconds() * 10 ** 7  # in 100*nanosecond
			if UUID(latest_payload["data"]["id"]).time - ws_state.last_update_ts > max_delta:
				ws_state.scope.cancel()

		await trio.sleep(poll_period.total_seconds())


def _build_notification_md(info: NotifyInfo) -> str:
	# TODO escape thread title?
	return "\n\n".join([
		f"from [{info.author}](/user/{info.author}) in [{info.thread_title}](/live/{info.thread_id})",
		"\n".join(">" + line for line in info.update_body.splitlines()),
		(
			f"[^^permalink](/live/{info.thread_id}/updates/{info.update_id})"
			+ f" [^^context](/live/{info.thread_id}?after=LiveUpdate_{info.update_id})"
			+ " [^^about ^^these ^^notifications](/r/live_mentions/wiki)"
		),
	])


@dataclass
class RedditNotifier(Notifier):
	token_ref: TokenRef
	logger: Logger

	def notify(self, info: NotifyInfo) -> None:
		params = {
			"raw_json": "1",
			"api_type": "json",
			"subject": "username mention",
			"text": _build_notification_md(info),
		}
		req = Request(
			"https://oauth.reddit.com/api/compose",
			method = "POST",
			data = urlencode({**params, "to": info.recipient}).encode("ascii"),
			headers = self.token_ref.get_common_headers(),
		)
		resp = try_request(req, self.logger)
		if resp:
			with resp:
				data = json.load(resp)["json"]
			if data["errors"]:
				[(code, _, _)] = data["errors"]
				if code != "USER_DOESNT_EXIST":
					raise RuntimeError("PM send error: {code}")


@dataclass
class RedditThreadTitleProvider(ThreadTitleProvider):
	token_ref: TokenRef

	def title_for(self, thread_id: str) -> str:
		req = Request(
			f"https://oauth.reddit.com/live/{thread_id}/about?raw_json=1",
			headers = self.token_ref.get_common_headers()
		)
		with urlopen(req) as resp:
			return json.load(resp)["data"]["title"]


@dataclass
class RedditUpdateChecker(UpdateChecker):
	token_ref: TokenRef
	logger: Logger

	def exists(self, thread_id: str, update_id: str) -> bool:
		req = Request(
			f"https://oauth.reddit.com/live/{thread_id}/updates/{update_id}?raw_json=1",
			headers = self.token_ref.get_common_headers(),
		)
		try:
			resp = try_request(req, self.logger, suppress_codes = [429, 500, 503])
		except HTTPError as e:
			e.read()
			e.close()
			if e.code == 404:
				return False
			else:
				raise
		else:
			if resp:
				resp.read()
				resp.close()
			return True


@dataclass
class WsConnectionEventStream(EventStream):
	EXCEPTION_TYPES: ClassVar[set[Type[BaseException]]] = {HandshakeError, ConnectionClosed}

	ws_conn: WebSocketConnection

	async def get_event(self) -> Event:
		while True:
			message = json.loads(await self.ws_conn.get_message())
			if message["type"] == "update":
				return UpdateEvent(message["payload"]["data"])
			elif message["type"] == "delete":
				return DeleteEvent(message["payload"])


async def _main(
	auth_id: int,
	buffer_time: timedelta,
	db_conn,
	logger: Logger,
	max_ws_delay: timedelta,
	mention_limit: int,
	poll_period: timedelta,
	thread_id: str,
	token_period: timedelta,
) -> None:

	@asynccontextmanager
	async def event_stream_factory() -> Iterator[WsConnectionEventStream]:
		resp = urlopen(Request(
			f"https://oauth.reddit.com/live/{thread_id}/about?raw_json=1",
			headers = token_ref.get_common_headers(),
		))
		with resp:
			ws_url = json.load(resp)["data"]["websocket_url"]

		async with open_websocket_url(ws_url) as ws_conn:
			yield WsConnectionEventStream(ws_conn)

	async with open_nursery() as nursery:
		token_ref = await nursery.start(_token_update_impl, token_period, db_conn, auth_id)

		ws_state = WsListenerState()
		(event_tx, event_rx) = open_memory_channel(0)
		nursery.start_soon(
			ws_listener_impl,
			event_stream_factory, WsConnectionEventStream.EXCEPTION_TYPES, ws_state, event_tx, logger.getChild("WS")
		)
		nursery.start_soon(
			_api_poll_impl,
			ws_state, token_ref, thread_id, event_tx, poll_period, max_ws_delay, logger.getChild("poll")
		)

		notifier = RedditNotifier(token_ref, logger)
		title_provider = RedditThreadTitleProvider(token_ref)
		deletion_checker = RedditUpdateChecker(token_ref, logger)
		nursery.start_soon(
			message_sender_impl,
			notifier, title_provider, deletion_checker, ws_state, thread_id, event_rx, buffer_time, mention_limit,
			logger.getChild("sender")
		)


if __name__ == "__main__":
	arg_parser = ArgumentParser(__package__)
	arg_parser.add_argument("config_path")
	args = arg_parser.parse_args()

	config_parser = ConfigParser()
	with open(args.config_path) as config_file:
		config_parser.read_file(config_file)

	main_cfg = config_parser["config"]

	auth_id = main_cfg.getint("auth ID")
	assert auth_id is not None

	token_period = timedelta(minutes = main_cfg.getint("token update period"))

	thread_id = main_cfg["thread ID"]

	mention_limit = main_cfg.getint("mention limit")
	assert mention_limit is not None and mention_limit >= 0

	buffer_time = timedelta(seconds = main_cfg.getfloat("message buffer time"))

	poll_period = timedelta(seconds = main_cfg.getfloat("API poll period"))

	max_ws_delay = timedelta(seconds = main_cfg.getfloat("max WS delay"))

	db_conn = psycopg2.connect(**config_parser["db connect params"])
	db_conn.autocommit = True

	logger = getLogger(__package__)
	logger.setLevel(NOTSET - 1)  # filter in handler(s) instead

	handler = StreamHandler(stdout)
	handler.setLevel(logging.DEBUG)
	handler.setFormatter(Formatter("{asctime:23}: {name:17}: {levelname:8}: {message}", style = "{"))
	logger.addHandler(handler)

	trio.run(
		_main,
		auth_id, buffer_time, db_conn, logger, max_ws_delay, mention_limit, poll_period, thread_id, token_period
	)
