from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from logging import getLogger
from numbers import Real
from os import environ
from typing import Iterator, TypeVar
from unittest import TestCase
from urllib.request import Request
from uuid import UUID, uuid1
import logging

from trio import CancelScope, MemoryReceiveChannel, move_on_after, open_memory_channel, open_nursery, sleep_until
from trio.testing import MockClock, wait_all_tasks_blocked
import trio

from . import message_sender_impl, TokenRef, try_request, WsListenerState, ws_listener_impl
from .abc import DeleteEvent, Event, EventStream, Notifier, NotifyInfo, ThreadTitleProvider, UpdateChecker, UpdateEvent


T = TypeVar("T")


@dataclass
class MockNotifier(Notifier):
	notifies: list[NotifyInfo] = field(default_factory = list)

	def notify(self, info: NotifyInfo) -> None:
		self.notifies.append(info)


class MockTitleProvider(ThreadTitleProvider):
	def title_for(self, thread_id: str) -> str:
		return thread_id


class UnusedUpdateChecker(UpdateChecker):
	def exists(self, thread_id: str, update_id: str) -> bool:
		raise AssertionError("unexpected deletion check")


@dataclass
class DictBackedUpdateChecker(UpdateChecker):
	exists_by_id: dict[str, bool] = field(default_factory = dict)
	queries: list[tuple[str, str]] = field(default_factory = list)

	def exists(self, thread_id: str, update_id: str) -> bool:
		self.queries.append((thread_id, update_id))
		return self.exists_by_id[update_id]


@dataclass
class ChannelBackedEventStream(EventStream):
	event_rx: MemoryReceiveChannel

	async def get_event(self) -> Event:
		obj = await self.event_rx.receive()
		if isinstance(obj, BaseException):
			raise obj
		else:
			return obj


@asynccontextmanager
async def null_async_context(val: T) -> Iterator[T]:
	yield val


def _update_event(author: str, mentions: list[str]) -> UpdateEvent:
	id_ = uuid1()
	body = str(id_) + " " + " ".join("/u/" + user for user in mentions)
	body_html = str(id_) + " " + " ".join(
		f'<a href="/u/{user}">/u/{user}</a>'
		for user in mentions
	)
	return UpdateEvent(
		{
			"id": str(id_),
			"name": "LiveUpdate_" + str(id_),
			"author": author,
			"body": body,
			"body_html": body_html,
		}
	)


def setUpModule() -> None:
	getLogger().setLevel(logging.CRITICAL + 1)  # disable logging


class TaskTests(TestCase):
	@contextmanager
	def _assert_completes_in(self, duration_seconds: Real) -> Iterator[None]:
		with move_on_after(duration_seconds) as scope:
			yield
		if scope.cancelled_caught:
			self.fail("operation timed out")

	def test_buffering(self) -> None:
		"""
		No API poll task, just a custom task mocking the WS listener by playing events over a channel, and a real
		notifier task injected with mocks to capture output. The WS state is hard-coded as up, so no testing of recovery
		from WS connection issues is included in this test.

		Primarily it makes sure that updates are passed through the buffer and processed if and only if a delete event
		doesn't arrive while they're buffered.
		"""

		buffer_time = timedelta(seconds = 1)
		notifier = MockNotifier()
		(event_tx, event_rx) = open_memory_channel(0)

		ign_event = _update_event("sender", [])
		event_a = _update_event("sender", ["rec1", "rec2"])
		event_b = _update_event("sender", ["rec1", "rec2"])

		async def event_stream_impl(scope: CancelScope) -> None:
			await sleep_until(1.0)
			await event_tx.send(ign_event)
			await event_tx.send(event_a)

			await sleep_until(1.5)
			self.assertFalse(notifier.notifies)
			await event_tx.send(DeleteEvent(event_a.payload_data["id"]))

			await sleep_until(2.0)
			await event_tx.send(event_b)

			# let things settle and terminate both tasks
			await sleep_until(4.0)
			scope.cancel()

		async def main() -> None:
			async with open_nursery() as nursery:
				nursery.start_soon(event_stream_impl, nursery.cancel_scope)
				nursery.start_soon(
					message_sender_impl,
					notifier, MockTitleProvider(), UnusedUpdateChecker(), WsListenerState(down = False), "ThreadId",
					event_rx, buffer_time, getLogger()
				)

		trio.run(main, clock = MockClock(autojump_threshold = 0))

		# only notifications sent are for event B
		self.assertEqual(len(notifier.notifies), 2)
		self.assertEqual({i.recipient for i in notifier.notifies}, {"rec1", "rec2"})
		self.assertTrue(all(i.author == "sender" for i in notifier.notifies))
		self.assertTrue(all(i.update_id == event_b.payload_data["id"] for i in notifier.notifies))

	def test_ws_trouble(self) -> None:
		"""
		Ensure that the notifier task responds correctly to the WS connection going down and coming back up.
		"""

		thread_id = "ThreadId"
		buffer_time = timedelta(seconds = 3)
		update_checker = DictBackedUpdateChecker()
		notifier = MockNotifier()
		(event_tx, event_rx) = open_memory_channel(0)

		# sent while WS is up, checked and notifies sent
		event_a = _update_event("sender", ["rec"])

		# sent and deleted while WS is down, checked and discarded from buffer
		event_b = _update_event("sender", ["rec"])

		# sent on WS reconnect, not checked and notifies sent
		event_c = _update_event("sender", ["rec"])

		# sent after WS reconnect, not checked and notifies sent
		event_d = _update_event("sender", ["rec"])

		ws_state = WsListenerState(down = False)

		async def ws_simulator_impl() -> None:
			update_checker.exists_by_id[event_a.payload_data["id"]] = True
			await event_tx.send(event_a)

			await sleep_until(1.0)
			ws_state.down = True
			update_checker.exists_by_id[event_b.payload_data["id"]] = True
			await event_tx.send(event_b)

			await sleep_until(2.0)
			update_checker.exists_by_id[event_b.payload_data["id"]] = False
			ws_state.down = False
			ws_state.resync_ts = UUID(event_c.payload_data["id"]).time
			await event_tx.send(event_c)

			await sleep_until(3.0)
			await event_tx.send(event_d)

		async def main() -> None:
			async with open_nursery() as nursery:
				nursery.start_soon(ws_simulator_impl)
				nursery.start_soon(
					message_sender_impl,
					notifier, MockTitleProvider(), update_checker, ws_state, "ThreadId", event_rx, buffer_time,
					getLogger()
				)

				await sleep_until(3.5)
				self.assertEqual([i.update_id for i in notifier.notifies], [event_a.payload_data["id"]])
				notifier.notifies.clear()
				self.assertEqual(update_checker.queries, [(thread_id, event_a.payload_data["id"])])
				update_checker.queries.clear()

				await sleep_until(4.5)
				self.assertFalse(notifier.notifies)
				self.assertEqual(update_checker.queries, [(thread_id, event_b.payload_data["id"])])
				update_checker.queries.clear()

				await sleep_until(5.5)
				self.assertEqual([i.update_id for i in notifier.notifies], [event_c.payload_data["id"]])
				notifier.notifies.clear()
				self.assertFalse(update_checker.queries)

				await sleep_until(6.5)
				self.assertEqual([i.update_id for i in notifier.notifies], [event_d.payload_data["id"]])
				notifier.notifies.clear()
				self.assertFalse(update_checker.queries)

				nursery.cancel_scope.cancel()

		trio.run(main, clock = MockClock(autojump_threshold = 0))

	def test_ws_listener_cancel(self) -> None:
		ws_state = WsListenerState()

		(event_in_tx, event_in_rx) = open_memory_channel(0)
		event_stream = ChannelBackedEventStream(event_in_rx)
		event_stream_factory = lambda: null_async_context(event_stream)

		(event_out_tx, event_out_rx) = open_memory_channel(0)

		async def main():
			async with open_nursery() as nursery:
				nursery.start_soon(
					ws_listener_impl,
					event_stream_factory, set(), ws_state, event_out_tx, getLogger()
				)

				event_1 = _update_event("author", [])
				await event_in_tx.send(event_1)
				with self._assert_completes_in(1):
					event_out = await event_out_rx.receive()
				self.assertIs(event_out, event_1)

				event_2 = _update_event("author", ["recipient"])
				await event_in_tx.send(event_2)
				with self._assert_completes_in(1):
					event_out = await event_out_rx.receive()
				self.assertIs(event_out, event_2)

				initial_scope = ws_state.scope
				initial_scope.cancel()
				await wait_all_tasks_blocked()
				self.assertIsNotNone(ws_state.scope)
				self.assertIsNot(ws_state.scope, initial_scope)
				self.assertTrue(ws_state.down)

				event_3 = _update_event("author", [])
				ts = UUID(event_3.payload_data["id"]).time
				await event_in_tx.send(event_3)
				with self._assert_completes_in(1):
					event_out = await event_out_rx.receive()
				self.assertIs(event_out, event_3)
				self.assertEqual(ws_state.last_update_ts, ts)
				self.assertFalse(ws_state.down)
				self.assertEqual(ws_state.resync_ts, ts)

				nursery.cancel_scope.cancel()

		trio.run(main, clock = MockClock(autojump_threshold = 0))

	def test_ws_listener_error(self) -> None:
		"""Test that the WS listener responds correctly to errors from the input event stream."""

		class MockStreamException(Exception):
			pass

		ws_state = WsListenerState()

		(event_in_tx, event_in_rx) = open_memory_channel(0)
		event_stream = ChannelBackedEventStream(event_in_rx)
		event_stream_factory = lambda: null_async_context(event_stream)

		(event_out_tx, event_out_rx) = open_memory_channel(0)

		async def main():
			async with open_nursery() as nursery:
				nursery.start_soon(
					ws_listener_impl,
					event_stream_factory, {MockStreamException}, ws_state, event_out_tx, getLogger()
				)

				event_1 = _update_event("author", [])
				await event_in_tx.send(event_1)
				with self._assert_completes_in(1):
					event_out = await event_out_rx.receive()
				self.assertIs(event_out, event_1)

				initial_scope = ws_state.scope
				await event_in_tx.send(MockStreamException())
				await wait_all_tasks_blocked()
				self.assertIsNotNone(ws_state.scope)
				self.assertIsNot(ws_state.scope, initial_scope)
				self.assertTrue(ws_state.down)

				event_2 = _update_event("author", ["recipient"])
				ts = UUID(event_2.payload_data["id"]).time
				await event_in_tx.send(event_2)
				with self._assert_completes_in(1):
					event_out = await event_out_rx.receive()
				self.assertIs(event_out, event_2)
				self.assertEqual(ws_state.last_update_ts, ts)
				self.assertFalse(ws_state.down)
				self.assertEqual(ws_state.resync_ts, ts)

				nursery.cancel_scope.cancel()

		trio.run(main, clock = MockClock(autojump_threshold = 0))


class RedditTests(TestCase):
	"""Tests for interactions with Reddit."""

	@classmethod
	def setUpClass(cls) -> None:
		cls._TOKEN_REF = TokenRef(environ["API_TOKEN"])

	def test_try_request(self) -> None:
		self.assertIsNotNone(
			try_request(
				Request(
					"https://oauth.reddit.com/api/live/happening_now",
					headers = self._TOKEN_REF.get_common_headers(),
				),
				getLogger(),
			)
		)

	def test_try_request_http_error(self) -> None:
		self.assertIsNone(
			try_request(
				Request("https://oauth.reddit.com/api/live/happening_now"),
				getLogger(),
				suppress_codes = [403],
			)
		)
